#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
from datetime import datetime, date

from firefly.presentation.web.js_libs.mithril import m
from firefly.presentation.web.js_libs.inflection import inflection
from firefly.presentation.web.js_libs.moment import moment
from firefly.presentation.web.polyfills import MissingType, ValueObject, Entity
from firefly import Validator, IsOneOf

import firefly.domain as ffd

# __pragma__('kwargs')


def form(entity: ffd.Entity, form_config: dict = None):
    dto = {}
    errors = {}
    fields = {}
    validator = Validator()
    config = {
        'placeholders': True,
        'labels': False,
        'id': 'form',
    }
    config.update(form_config or {})

    for field_ in dir(entity):
        if field_.startswith('__') or callable(getattr(entity, field_)):
            continue
        if field_ in config.get('skip', []):
            continue
        value = getattr(entity, field_)
        fields[field_] = value
        if isinstance(value, dict) and 'default_factory' in value:
            if callable(value['default_factory']):
                value = value['default_factory']()
                if isinstance(value, ffd.Empty):
                    value = ''
            else:
                value = value['default']
        if isinstance(value, (MissingType, bool)) or value is None:
            value = ''

        if field_ == 'name' and value == 'cls':
            value = ''
        dto[field_] = value

    def handle_submit(event):
        event.preventDefault()
        if 'onsubmit' in config:
            config['onsubmit'](dto)

    def get_type(f: str):
        if 'types' in config and f in config['types']:
            return config['types'][f]

        if 'metadata' in fields[f] and 'type' in fields[f]['metadata']:
            return fields[f]['metadata']['type']

        if 'metadata' in fields[f] and 'py_metatype' in fields[f]['metadata']:
            return fields[f]['metadata']['py_metatype']

        return None

    def validate():
        nonlocal errors
        e = validator.validate(dto, entity.__class__)
        if e['count'] > 0:
            errors = e['errors']

    def humanize(val: str):
        return inflection.titleize(inflection.humanize(val))

    def has_errors(f: str):
        return f in errors and len(errors[f]) > 0 and has_value(f)

    def is_required(f: str):
        return 'metadata' in fields[f] and fields[f]['metadata'].get('required', False) \
               and (not config.get('nested', False) or config.get('required', False))

    def label(f: str, conf=None):
        return f'label', m(f'label[for="{f}"].block', conf, humanize(f))

    def error_container(f: str, conf=None):
        return m(f'span[id="{f}_error"].errors', conf, error_messages(f))

    def error_messages(f: str):
        if f not in errors or not has_value(f):
            return []
        return m('ul', map(lambda msg: m('li', msg), errors[f]))

    def field_set(fields_: list, options: dict, legend: str):
        children = []
        if legend is not None:
            children.append(m('legend', legend))
        children.extend(list(map(field_group, fields_)))

        return m(f'fieldset', options, children)

    def field_group(field_: str):
        t = get_type(field_)
        nested_form = issubclass(t, (ValueObject, Entity))
        children = []
        if config['labels']:
            children.append(label(field_))

        if nested_form:
            children.append = [m(Form(t(), {'nested': True, 'required': is_required(field_)}))]
        else:
            children.append(form_field(field_))

        if nested_form is False:
            children.append(error_container(field_))

        classes = ''
        if has_errors(field_):
            classes += '.error'

        if nested_form:
            return children
        else:
            return m(f'div.field-group{classes}', children)

    def enum_fields(f: str):
        if 'metadata' in fields[f] and 'validators' in fields[f]['metadata']:
            validators = fields[f]['metadata']['validators']
        else:
            return None

        for validator in validators:
            if isinstance(validator, IsOneOf):
                values = validator.values
                if callable(values):
                    values = values(dto)
                return values

        return None

    def form_field(f: str, conf=None, children=None):
        conf = conf or {}
        if is_required(f):
            conf['required'] = 'required'

        t = get_type(f)

        def set_value(e):
            if t is bool:
                dto[f] = e.target.checked
            else:
                dto[f] = e.target.value
            validate()

        conf['oninput'] = set_value

        classes = ''
        if has_value(f) or t is bool:
            classes += '.has-value'

        if t is datetime:
            return datetime_inputs(f, classes, conf)
        elif t is date:
            return date_input(f, classes, conf)
        elif t is bool:
            return boolean_input(f, classes, conf)
        elif t is int or t is float:
            return text_input(f, classes, conf, children, input_type='number')

        values = enum_fields(f)
        if values is not None:
            return select(f, values, classes, conf)

        return text_input(f, classes, conf, children)

    def boolean_input(f: str, classes: str, conf: dict):
        if dto[f]:
            conf['checked'] = 'checked'
        return m(f'input[type="checkbox"][id="{f}"][name="{f}"].form-checkbox-input{classes}', conf)

    def text_input(f: str, classes: str, conf: dict, children, input_type: str = 'text'):
        return m(
            f'input[type="{input_type}"][id="{f}"][name="{f}"][value="{dto[f]}"][placeholder="{humanize(f)}"]'
            f'.form-text-input.px-2.py-1.rounded-sm{classes}.h-12',
            conf,
            children
        )

    def select(f: str, options, classes: str, conf: dict):
        value = dto[f]
        options = list(options)
        options.insert(0, humanize(f))

        def option(f):
            selected = ''
            if f == value:
                selected = '[selected="selected"]'
            return m(f'option[value="{f}"]{selected}', f)

        return m(f'select.form-select{classes}.h-12', conf, list(map(option, options)))

    def date_input(f: str, classes: str, conf: dict):
        return m(
            f'input[type="date"][id="{f}"][name="{f}"]'
            f'[value="{moment(dto[f]).format("YYYY-MM-DD")}"]'
            f'.form-date-input.px-2.py-1.rounded-sm.h-12{classes}',
            conf,
        )

    def datetime_inputs(f: str, classes: str, conf: dict):
        return [
            m(
                f'input[type="date"][id="{f}_date"][name="{f}_date"]'
                f'[value="{moment(dto[f]).format("YYYY-MM-DD")}"]'
                f'.form-date-input.px-2.py-1.rounded-sm.h-12{classes}',
                conf,
            ),
            m(
                f'input[type="time"][id="{f}_date"][name="{f}_date"]'
                f'[value="{moment(dto[f]).format("HH:mm")}"]'
                f'.form-time-input.px-2.py-1.rounded-sm.h-12{classes}',
                conf,
            ),
        ]

    def has_value(f: str):
        return dto[f] is not None and (not isinstance(dto[f], str) or dto[f] != '')

    def _form(conf=None, children=None):
        if 'nested' in config and config['nested']:
            return children

        return m(
            'div.form-container.flex.flex-col',
            m(
                f'form#{config["id"]}',
                conf,
                children
            )
        )

    def view():
        config = {'onsubmit': handle_submit}
        if 'fields' in config:
            keys = config['fields']
        else:
            keys = dto.keys()

        if 'exclude_fields' in config:
            keys = [x for x in keys if x not in config['exclude_fields']]

        if 'fieldsets' in config:
            form_fields = []
            for fieldset in config['fieldsets']:
                form_fields.append(
                    field_set(
                        fieldset['fields'],
                        fieldset['options'],
                        fieldset['legend'] if 'legend' in fieldset else None
                    )
                )
        else:
            form_fields = list(map(field_group, keys))

        if 'nested' not in config or not config['nested']:
            form_fields.append(m('input[type="submit"][value="Submit"].hidden.md:block'))

        return _form(config, form_fields)

    return {
        'view': view
    }


# class Form:
#     """
#     TODO
#         - Entity/VO collections
#         - Validate on submit
#         - More types? (int, float, email, phone number)...
#     """
#     def __init__(self, entity: ffd.Entity, config: dict = None):
#         self._entity = entity
#         self._dto = {}
#         self._errors = {}
#         self._config = {
#             'placeholders': True,
#             'labels': False,
#             'id': 'form',
#         }
#         self._config.update(config or {})
#         self._fields = {}
#         self._validator = Validator()
#
#         for field_ in dir(self._entity):
#             if field_.startswith('__') or callable(getattr(self._entity, field_)):
#                 continue
#             if field_ in self._config.get('skip', []):
#                 continue
#             value = getattr(self._entity, field_)
#             self._fields[field_] = value
#             if isinstance(value, dict) and 'default_factory' in value:
#                 if callable(value['default_factory']):
#                     value = value['default_factory']()
#                     if isinstance(value, ffd.Empty):
#                         value = ''
#                 else:
#                     value = value['default']
#             if isinstance(value, (MissingType, bool)) or value is None:
#                 value = ''
#
#             if field_ == 'name' and value == 'cls':
#                 value = ''
#             self._dto[field_] = value
#
#     def validate(self):
#         errors = self._validator.validate(self._dto, self._entity.__class__)
#         if errors['count'] > 0:
#             self._errors = errors['errors']
#
#     @staticmethod
#     def _wrap(component: str, vnode):
#         return vnode
#
#     @staticmethod
#     def _humanize(val: str):
#         return inflection.titleize(inflection.humanize(val))
#
#     def _has_errors(self, field_: str):
#         return field_ in self._errors and len(self._errors[field_]) > 0 and self._has_value(field_)
#
#     def _is_required(self, field_: str):
#         return 'metadata' in self._fields[field_] and self._fields[field_]['metadata'].get('required', False) \
#                and (not self._config.get('nested', False) or self._config.get('required', False))
#
#     def _form(self, config=None, children=None):
#         if 'nested' in self._config and self._config['nested']:
#             return children
#
#         return self._wrap(
#             'form',
#             m(
#                 'div.form-container.flex.flex-col',
#                 m(
#                     f'form#{self._config["id"]}',
#                     config,
#                     children
#                 )
#             )
#         )
#
#     def _field_set(self, fields: list, options: dict, legend: str):
#         children = []
#         if legend is not None:
#             children.append(m('legend', legend))
#         children.extend(list(map(self._field_group, fields)))
#
#         return self._wrap('field_set', m(f'fieldset', options, children))
#
#     def _field_group(self, field_: str):
#         t = self._get_type(field_)
#         nested_form = issubclass(t, (ValueObject, Entity))
#         children = []
#         if self._config['labels']:
#             children.append(self._label(field_))
#
#         if nested_form:
#             children.append = [m(Form(t(), {'nested': True, 'required': self._is_required(field_)}))]
#         else:
#             children.append(self._field(field_))
#
#         if nested_form is False:
#             children.append(self._error_container(field_))
#
#         classes = ''
#         if self._has_errors(field_):
#             classes += '.error'
#
#         if nested_form:
#             return self._wrap('field_group', children)
#         else:
#             return self._wrap('field_group', m(f'div.field-group{classes}', children))
#
#     def _label(self, field_: str, config=None):
#         return self._wrap(f'label', m(f'label[for="{field_}"].block', config, self._humanize(field_)))
#
#     def _error_container(self, field_: str, config=None):
#         return self._wrap('error', m(f'span[id="{field_}_error"].errors', config, self._error_messages(field_)))
#
#     def _error_messages(self, field_: str):
#         if field_ not in self._errors or not self._has_value(field_):
#             return []
#         return m('ul', map(lambda msg: m('li', msg), self._errors[field_]))
#
#     def _get_type(self, field_: str):
#         if 'types' in self._config and field_ in self._config['types']:
#             return self._config['types'][field_]
#
#         if 'metadata' in self._fields[field_] and 'type' in self._fields[field_]['metadata']:
#             return self._fields[field_]['metadata']['type']
#
#         if 'metadata' in self._fields[field_] and 'py_metatype' in self._fields[field_]['metadata']:
#             return self._fields[field_]['metadata']['py_metatype']
#
#         return None
#
#     def _enum_fields(self, field_: str):
#         if 'metadata' in self._fields[field_] and 'validators' in self._fields[field_]['metadata']:
#             validators = self._fields[field_]['metadata']['validators']
#         else:
#             return None
#
#         for validator in validators:
#             if isinstance(validator, IsOneOf):
#                 values = validator.values
#                 if callable(values):
#                     values = values(self._dto)
#                 return values
#
#         return None
#
#     def _field(self, field_: str, config=None, children=None):
#         config = config or {}
#         if self._is_required(field_):
#             config['required'] = 'required'
#
#         t = self._get_type(field_)
#
#         def set_value(e):
#             if t is bool:
#                 self._dto[field_] = e.target.checked
#             else:
#                 self._dto[field_] = e.target.value
#             self.validate()
#         config['oninput'] = set_value
#
#         classes = ''
#         if self._has_value(field_) or t is bool:
#             classes += '.has-value'
#
#         if t is datetime:
#             return self._wrap('field', self._datetime_inputs(field_, classes, config))
#         elif t is date:
#             return self._wrap('field', self._date_input(field_, classes, config))
#         elif t is bool:
#             return self._wrap('field', self._boolean_input(field_, classes, config))
#         elif t is int or t is float:
#             return self._wrap('field', self._text_input(field_, classes, config, children, input_type='number'))
#
#         values = self._enum_fields(field_)
#         if values is not None:
#             return self._wrap('field', self._select(field_, values, classes, config))
#
#         return self._wrap('field', self._text_input(field_, classes, config, children))
#
#     def _boolean_input(self, field_: str, classes: str, config: dict):
#         if self._dto[field_]:
#             config['checked'] = 'checked'
#         return m(f'input[type="checkbox"][id="{field_}"][name="{field_}"].form-checkbox-input{classes}', config)
#
#     def _text_input(self, field_: str, classes: str, config: dict, children, input_type: str = 'text'):
#         return m(
#             f'input[type="{input_type}"][id="{field_}"][name="{field_}"][value="{self._dto[field_]}"][placeholder="{self._humanize(field_)}"]'
#             f'.form-text-input.px-2.py-1.rounded-sm{classes}.h-12',
#             config,
#             children
#         )
#
#     def _select(self, field_: str, options, classes: str, config: dict):
#         value = self._dto[field_]
#         options = list(options)
#         options.insert(0, self._humanize(field_))
#
#         def option(f):
#             selected = ''
#             if f == value:
#                 selected = '[selected="selected"]'
#             return m(f'option[value="{f}"]{selected}', f)
#
#         return m(f'select.form-select{classes}.h-12', config, list(map(option, options)))
#
#     def _date_input(self, field_: str, classes: str, config: dict):
#         return m(
#             f'input[type="date"][id="{field_}"][name="{field_}"]'
#             f'[value="{moment(self._dto[field_]).format("YYYY-MM-DD")}"]'
#             f'.form-date-input.px-2.py-1.rounded-sm.h-12{classes}',
#             config,
#         )
#
#     def _datetime_inputs(self, field_: str, classes: str, config: dict):
#         return [
#             m(
#                 f'input[type="date"][id="{field_}_date"][name="{field_}_date"]'
#                 f'[value="{moment(self._dto[field_]).format("YYYY-MM-DD")}"]'
#                 f'.form-date-input.px-2.py-1.rounded-sm.h-12{classes}',
#                 config,
#             ),
#             m(
#                 f'input[type="time"][id="{field_}_date"][name="{field_}_date"]'
#                 f'[value="{moment(self._dto[field_]).format("HH:mm")}"]'
#                 f'.form-time-input.px-2.py-1.rounded-sm.h-12{classes}',
#                 config,
#             ),
#         ]
#
#     def _has_value(self, field_: str):
#         return self._dto[field_] is not None and (not isinstance(self._dto[field_], str) or self._dto[field_] != '')
#
#     def _handle_submit(self, event):
#         event.preventDefault()
#         if 'onsubmit' in self._config:
#             self._config['onsubmit'](self._dto)
#
#     def view(self):
#         config = {'onsubmit': self._handle_submit}
#         if 'fields' in self._config:
#             keys = self._config['fields']
#         else:
#             keys = self._dto.keys()
#
#         if 'exclude_fields' in self._config:
#             keys = [x for x in keys if x not in self._config['exclude_fields']]
#
#         if 'fieldsets' in self._config:
#             form_fields = []
#             for fieldset in self._config['fieldsets']:
#                 form_fields.append(
#                     self._field_set(
#                         fieldset['fields'],
#                         fieldset['options'],
#                         fieldset['legend'] if 'legend' in fieldset else None
#                     )
#                 )
#         else:
#             form_fields = list(map(self._field_group, keys))
#
#         if 'nested' not in self._config or not self._config['nested']:
#             form_fields.append(m('input[type="submit"][value="Submit"].hidden.md:block'))
#
#         return self._form(config, form_fields)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from dataclasses import asdict
from typing import List

import firefly as ff

import firefly_test.calendar.domain as cal


class Calendar(ff.AggregateRoot):
    id: str = ff.id_()
    events: List[cal.Event] = ff.list_()

    def add_event(self, event: cal.Event):
        self.events.append(event)
        self.dispatch('EventAdded', asdict(event))

    def add_reminder(self, event_id: str, reminder: cal.Reminder):
        for event in self.events:
            if event.id == event_id:
                event.reminders.append(reminder)
                self.dispatch('ReminderAdded', asdict(reminder))



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List, Callable

from firefly.domain.service.messaging.middleware import Middleware
import firefly.domain as ffd

# __pragma__('kwargs')
# __pragma__('opov')


class MiddlewareStack:
    def __init__(self, middleware: List[Middleware]):
        self._middleware = middleware

    @property
    def middleware(self):
        return self._middleware

    @middleware.setter
    def middleware(self, value: List[Middleware]):
        self._middleware = value

    def add(self, item: Middleware):
        self._middleware.append(item)

    def insert(self, index: int, item: Middleware):
        self._middleware.insert(index, item)

    def replace(self, which: type, with_: ffd.Middleware):
        for i, mw in enumerate(self.middleware):
            if mw.__class__ is which:
                self.middleware[i] = with_
                return True
        return False

    def __call__(self, msg: ffd.Message):
        def cb(message, *args, **kwargs):
            return message

        for m in reversed(self._middleware):
            cb = self._nest(cb, m)

        return cb(msg)

    @staticmethod
    def _nest(cb: Callable, middleware: Middleware):
        def callback(message, next_=None):
            return middleware(message, next_=cb)

        return callback



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from firefly.domain import error


class PreDeployHook:
    def __call__(self, for_: str):
        def agent_wrapper(cls):
            try:
                cls.set_agent_extension(for_, 'add_pre_deployment_hook')
            except AttributeError:
                raise error.FrameworkError('@agent used on invalid target')
            return cls

        return agent_wrapper


class PostDeployHook:
    def __call__(self, for_: str):
        def agent_wrapper(cls):
            try:
                cls.set_agent_extension(for_, 'add_post_deployment_hook')
            except AttributeError:
                raise error.FrameworkError('@agent used on invalid target')
            return cls

        return agent_wrapper


class Agent:
    pre_deploy_hook: PreDeployHook = PreDeployHook()
    post_deploy_hook: PostDeployHook = PostDeployHook()

    def __call__(self, name: str):
        def agent_wrapper(cls):
            try:
                cls.set_agent(name)
            except AttributeError:
                raise error.FrameworkError('@agent used on invalid target')
            return cls

        return agent_wrapper


agent = Agent()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import sqlite3
from typing import Type, Optional, Union

import firefly.domain as ffd
import inflection
from firefly.infrastructure.repository.rdb_repository import Column, Index

from .legacy_storage_interface import LegacyStorageInterface


class SqliteStorageInterface(LegacyStorageInterface, ffd.LoggerAware):
    _serializer: ffd.Serializer = None
    _sql_prefix = 'sqlite'
    _map_indexes = True

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._config = kwargs
        self._connection: Optional[sqlite3.Connection] = None

    def _disconnect(self):
        if self._connection is not None:
            self._connection.close()
            self._connection = None
            self._tables_checked = []

    def _ensure_connected(self):
        if self._connection is not None:
            return

        try:
            host = self._config['host']
        except KeyError:
            raise ffd.ConfigurationError(f'host is required in sqlite_storage_interface')

        self._connection = sqlite3.connect(host, detect_types=sqlite3.PARSE_DECLTYPES)
        self._connection.row_factory = sqlite3.Row

    @staticmethod
    def _fqtn(entity: Type[ffd.Entity]):
        return inflection.tableize(entity.get_fqn()).replace('.', '_')

    def _build_entity(self, entity: Type[ffd.Entity], data, raw: bool = False):
        if raw:
            return self._serializer.deserialize(data['document'])
        ret = entity.from_dict(self._load_relationships(entity, self._serializer.deserialize(data['document'])))
        setattr(ret, '__ff_version', data['version'])
        return ret

    def _get_table_columns(self, entity: Type[ffd.Entity]):
        ret = []
        results = self._execute(*self._generate_query(entity, f'{self._sql_prefix}/get_columns.sql'))

        for row in results:
            d = dict(row)
            ret.append(Column(name=d['name'], type=d['type']))

        return ret

    def _get_table_indexes(self, entity: Type[ffd.Entity]):
        ret = []
        results = self._execute(*self._generate_query(entity, f'{self._sql_prefix}/get_indexes.sql'))

        for row in results:
            d = dict(row)
            if d['sql'] is None:
                continue
            ret.append(Index(name=d['name']))

        return ret

    def _execute(self, sql: str, params: dict = None):
        self._ensure_connected()
        cursor = self._connection.cursor()
        self.info(sql)
        self.info(params)
        cursor.execute(sql, params)
        self._connection.commit()

        if sql.startswith('update ') or sql.startswith('insert '):
            return cursor.rowcount

        return cursor.fetchall()

    def create_schema(self, entity_type: Type[ffd.Entity]):
        return True



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import os
import pathlib
from os.path import dirname
from typing import List, Tuple

import firefly as ff


class LocalFileSystem(ff.FileSystem):
    _serializer: ff.Serializer = None

    def read(self, file_name: str):
        with open(file_name, 'rb') as fp:
            return ff.File(
                name=file_name,
                content=fp.read()
            )

    def write(self, file: ff.File, path: str = None):
        d = path or dirname(file.name)
        if not os.path.exists(d):
            pathlib.Path(d).mkdir(parents=True, exist_ok=True)
        mode = 'wb' if isinstance(file.content, bytes) else 'w'
        data = file.content
        if not isinstance(data, str):
            data = data.decode('utf-8')
        with open((path or '').rstrip('/') + '/' + file.name, mode) as fp:
            fp.write(data)

    def list(self, path: str) -> List[Tuple[str, dict]]:
        ret = []
        for (_, _, filenames) in os.walk(path):
            # TODO we should have file sizes and last_modified here
            ret.extend([
                (f, {}) for f in filenames
            ])

        return ret

    def filter(self, path: str, fields: list, criteria: ff.BinaryOp):
        if not path.lower().endswith('.json'):
            raise NotImplemented()

        file = self.read(path)
        data = self._serializer.deserialize(file.content)
        ret = []
        if not isinstance(data, list):
            data = [data]
        for d in data:
            if criteria.matches(d):
                ret.append(d)

        return ret



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List, Callable, Union, Tuple, Optional

import firefly.domain as ffd
import firefly.infrastructure as ffi
from firefly.domain.repository.repository import T

from .abstract_repository import AbstractRepository

DEFAULT_LIMIT = 999999999999999999


class DocumentRepository(AbstractRepository[T]):
    def __init__(self, interface: ffi.DocumentStorageInterface):
        super().__init__(interface=interface)

        self._entity_type = self._type()
        self._interface = interface
        self._interface._repository = self
        self._index = 0
        self._state = 'empty'
        self._query_details = {}

    def append(self, entity: Union[T, List[T], Tuple[T]], **kwargs):
        if not isinstance(entity, (list, tuple)):
            entity = [entity]

        for e in entity:
            if e not in self._entities:
                self._entities.append(e)
                self.debug('Entity added to repository: %s', str(e))
        self._state = 'partial'

    def remove(self, x: Union[T, List[T], Tuple[T], Callable, ffd.BinaryOp], **kwargs):
        if self._parent is not None:
            self._parent.remove(x)

        xs = x
        if not isinstance(x, (list, tuple)):
            xs = [x]

        for x in xs:
            self.debug('Entity removed from repository: %s', str(x))
            self._deletions.append(x)
            if isinstance(x, ffd.Entity) and x in self._entities:
                self._entities.remove(x)

    def find(self, x: Union[str, Callable, ffd.BinaryOp], **kwargs) -> T:
        ret = None
        if isinstance(x, str):
            entity = self._find_checked_out_entity(x)
            if entity is not None:
                return entity
            ret = self._interface.find(x, self._entity_type)
        else:
            if not isinstance(x, ffd.BinaryOp):
                x = self._get_search_criteria(x)
            results = self._interface.all(self._entity_type, x)
            if len(results) > 0:
                ret = results[0]

        if ret:
            self.register_entity(ret)
            if self._state == 'empty':
                self._state = 'partial'

        return ret

    def filter(self, x: Union[Callable, ffd.BinaryOp], **kwargs) -> DocumentRepository:
        self._query_details.update(kwargs)
        self._query_details['criteria'] = x

        return self.copy()

    def _do_filter(self, criteria: Union[Callable, ffd.BinaryOp], limit: int = None, offset: int = None,
                   raw: bool = False, sort: tuple = None) -> List[T]:
        if criteria is not None:
            criteria = self._get_search_criteria(criteria) if not isinstance(criteria, ffd.BinaryOp) else criteria
        if self._state == 'full':
            entities = list(filter(lambda e: criteria.matches(e), self._entities))
        else:
            entities = self._interface.all(
                self._entity_type, criteria=criteria, limit=limit, offset=offset, raw=raw, sort=sort
            )

            merged = []
            for entity in entities:
                if entity in self._entities:
                    merged.append(next(e for e in self._entities if e == entity))
                else:
                    merged.append(entity)
                    if raw is False:
                        self.register_entity(entity)
            if self._state == 'empty':
                self._state = 'partial'
            entities = merged
        return entities

    def sort(self, cb: Optional[Union[Callable, Tuple[Union[str, Tuple[str, bool]]]]] = None, **kwargs):
        if cb is None and 'key' in kwargs:
            return list(self).sort(**kwargs)

        if not isinstance(cb, tuple):
            cb = cb(ffd.EntityAttributeSpy(self._entity_type))
            if not isinstance(cb, tuple):
                cb = [(cb,)]
        self._query_details['sort'] = cb

        return self.copy()

    def clear(self):
        self._interface.clear(self._entity_type)

    def destroy(self):
        self._interface.destroy(self._entity_type)

    def copy(self):
        ret = self.__class__()
        ret._query_details = self._query_details.copy()
        ret._entities = []
        ret._entity_hashes = {}
        ret._deletions = []
        ret._parent = self

        deletions = self._deletions
        entities = self._new_entities()
        self.reset()
        self._deletions = deletions
        self._entities = entities

        return ret

    def __iter__(self):
        if 'raw' in self._query_details and self._query_details['raw'] is True:
            return iter(self._load_data())
        self._load_data()
        return iter(list(self._entities))

    def __len__(self):
        params = self._query_details.copy()
        if 'criteria' in params and not isinstance(params['criteria'], ffd.BinaryOp):
            params['criteria'] = self._get_search_criteria(params['criteria'])
        return self._interface.all(self._entity_type, count=True, **params)

    def __getitem__(self, item):
        if isinstance(item, slice):
            if item.start is not None:
                self._query_details['offset'] = item.start
            if item.stop is not None:
                self._query_details['limit'] = (item.stop - item.start) + 1
            else:
                self._query_details['limit'] = DEFAULT_LIMIT
        else:
            if len(self._entities) > item:
                return self._entities[item]
            self._query_details['offset'] = item
            self._query_details['limit'] = 1

        if 'raw' in self._query_details and self._query_details['raw'] is True:
            return self._load_data()

        self._load_data()

        if isinstance(item, slice):
            return self._entities
        elif len(self._entities) > 0:
            return self._entities[-1]

    def _load_data(self):
        query_details = self._query_details

        if 'criteria' not in query_details:
            query_details['criteria'] = None

        results = self._do_filter(**query_details)
        if 'raw' in query_details and query_details['raw'] is True:
            return results

        if isinstance(results, list):
            for entity in results:
                if entity not in self._entities:
                    self.register_entity(entity)

    def commit(self, force_delete: bool = False):
        self.debug('commit() called in %s', str(self))

        if len(self._deletions) > 0:
            self.debug('Deleting %s', self._deletions)
            self._interface.remove(self._deletions, force=force_delete)

        new_entities = self._new_entities()
        if len(new_entities) > 0:
            self.debug('Adding %s', new_entities)
            self._interface.add(new_entities)

        for entity in self._changed_entities():
            self.debug('Updating %s', entity)
            self._interface.update(entity)
        self.debug('Done in commit()')

    def __repr__(self):
        return f'RdbRepository[{self._entity_type}]'

    def _find_checked_out_entity(self, id_: str):
        for entity in self._entities:
            if entity.id_value() == id_:
                return entity

    def reset(self):
        super().reset()
        self._query_details = {}
        self._state = 'empty'



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import inspect

import firefly as ff
import firefly.domain.error as error
from firefly.domain.entity.core.cli_argument import CliArgument
from firefly.domain.entity.core.cli_endpoint import CliEndpoint


class Cli:
    def __call__(self, command: str, app: str = None, description: str = None, target: ff.TypeOfMessage = None,
                 alias: dict = None, help_: str = None, args_help: dict = None):
        def cli_wrapper(cls):
            arguments = []

            if inspect.isfunction(cls):
                args = ff.get_arguments(cls, none_type_unions=False)
            else:
                args = ff.get_arguments(cls.__call__, none_type_unions=False)

            for name, arg in args.items():
                arguments.append(CliArgument(
                    name=name,
                    type=arg['type'],
                    default=arg['default'] if arg['default'] is not inspect.Parameter.empty else None,
                    required=arg['default'] is inspect.Parameter.empty,
                    help=args_help[name] if isinstance(args_help, dict) and name in args_help else None,
                    alias=alias[name] if isinstance(alias, dict) and name in alias else None
                ))

            endpoint = CliEndpoint(
                app=app if app is not None else command.split(' ')[0],
                command=command,
                description=description,
                message=target,
                alias=alias,
                help=help_,
                arguments=arguments,
            )

            try:
                cls.add_endpoint(endpoint)
            except AttributeError:
                if inspect.isfunction(cls):
                    ff.add_endpoint(cls, endpoint)
                else:
                    raise error.FrameworkError('@cli used on invalid target')
            return cls

        return cli_wrapper


cli = Cli()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

# __pragma__('skip')
from abc import ABCMeta

from dataclasses import dataclass

# __pragma__('noskip')
# __pragma__('ecom')
"""?
from firefly.presentation.web.polyfills import ABCMeta, dataclass
?"""
# __pragma__('noecom')

import firefly.domain as ffd


# __pragma__('kwargs')
class MessageMeta(ABCMeta):
    def __new__(mcs, name, bases, dct, **kwargs):
        # __pragma__('skip')
        my_dict = dct.copy()
        # __pragma__('noskip')
        # __pragma__('ecom')
        """?
        my_dict = dct
        ?"""
        # __pragma__('noecom')

        if 'fields_' in kwargs and 'annotations_' in kwargs:
            for k, v in kwargs['fields_'].items():
                my_dict[k] = ffd.optional(default=v)
            if '__annotations__' in my_dict:
                my_dict['__annotations__'].update(kwargs['annotations_'])
            else:
                my_dict['__annotations__'] = kwargs['annotations_']

        ret = type.__new__(mcs, name, bases, my_dict)

        return dataclass(ret, eq=False, repr=False)

    def __instancecheck__(self, instance):
        if self is ffd.Message:
            return True
        return self in instance.__class__.__bases__
# __pragma__('nokwargs')



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import atexit
import errno
import importlib.util
import os
import signal
import subprocess
import sys
from typing import Optional

import firefly.domain as ffd
import firefly.infrastructure as ffi
import inflection


class DefaultAgent(ffd.ApplicationService, ffd.LoggerAware):
    _web_server: ffi.WebServer = None
    _config: ffd.Configuration = None
    _context_map: ffd.ContextMap = None
    _registry: ffd.Registry = None
    _transaction_handler: ffd.TransactionHandlingMiddleware = None

    def __init__(self):
        self._deployment: Optional[ffd.Deployment] = None

    def __call__(self, deployment: ffd.Deployment, start_server: bool = True, start_web_app: bool = False, **kwargs):
        self._deployment = deployment
        self._execute_ddl()

        if self._register_gateways not in self._web_server.extensions:
            self._web_server.add_extension(self._register_gateways)

        if start_web_app:
            self.info('Starting web app')
            self._start_web_app()

        if start_server:
            self._transaction_handler.reset_level()
            self.info('Starting web server')
            self._web_server.run()

    def _execute_ddl(self):
        for context in self._context_map.contexts:
            for entity in context.entities:
                if issubclass(entity, ffd.AggregateRoot) and entity is not ffd.AggregateRoot:
                    try:
                        repository = self._registry(entity)
                        if isinstance(repository, ffi.RdbRepository):
                            repository.migrate_schema()
                    except ffd.FrameworkError:
                        pass

    def _register_gateways(self, web_server: ffi.WebServer):
        for service in self._deployment.services:
            for api_gateway in service.api_gateways:
                for endpoint in api_gateway.endpoints:
                    route = endpoint.route
                    prefix = f'/{inflection.dasherize(service.name)}'
                    if not route.startswith(f'{prefix}/') and route != prefix:
                        route = f'{prefix}{route}'
                    web_server.add_endpoint(endpoint.method, route, endpoint.message)

    def _start_web_app(self):
        self._compile_web_app()
        cmd = './node_modules/.bin/webpack-dev-server -w --mode development --env local'
        webpack = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True, bufsize=0
        )

        def stop_webpack(a=None, b=None):
            try:
                os.killpg(webpack.pid, signal.SIGTERM)
            except ProcessLookupError:
                pass

        atexit.register(stop_webpack)
        signal.signal(signal.SIGTERM, stop_webpack)
        signal.signal(signal.SIGINT, stop_webpack)

        while True:
            output = webpack.stdout.readline()
            if output:
                print(output.decode().rstrip())
                if 'Compiled successfully' in output.decode() or 'Compiled with warnings' in output.decode():
                    break
            if webpack.poll() is not None:
                break
            sys.stdout.flush()

    def _compile_web_app(self):
        modules = []
        for name, context in self._config.contexts.items():
            module_name = f'{name}_web.admin'
            if importlib.util.find_spec(module_name) is not None:
                modules.append(module_name)
            module_name = f'{name}_web.app'
            if importlib.util.find_spec(module_name) is not None:
                modules.append(module_name)

        self._create_build_files(modules)
        self._transpile('build.entry')

    @staticmethod
    def _create_build_files(modules: list):
        try:
            os.mkdir('build')
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

        with open('build/app.py', 'w') as fp:
            for module in modules:
                if '_web.app' in module:
                    fp.write(f'import {module}\n')

        with open('build/admin.py', 'w') as fp:
            fp.write('import firefly_web.app\n')
            for module in modules:
                if '_web.admin' in module:
                    fp.write(f'import {module}\n')

        with open('build/entry.py', 'w') as fp:
            fp.write('import build.app\n')
            fp.write('import build.admin\n')

    def _transpile(self, module_name: str):
        cmd = f'. ./venv/bin/activate && ' \
              f'./venv/bin/python3 -m transcrypt --nomin --map --fcall --verbose "{module_name}"'
        self.info(os.system(cmd))



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from dataclasses import fields
from datetime import datetime
from typing import Type, get_type_hints, List, Union, Callable, Dict, Tuple

import firefly.domain as ffd
import inflection
# noinspection PyDataclass
from jinjasql import JinjaSql

from .rdb_repository import Index, Column


# noinspection PyDataclass
class AbstractStorageInterface(ffd.LoggerAware, ABC):
    _serializer: ffd.Serializer = None
    _registry: ffd.Registry = None
    _cache: dict = {}

    def disconnect(self):
        self._disconnect()

    def _disconnect(self):
        pass

    def add(self, entity: Union[ffd.Entity, List[ffd.Entity]]):
        self._check_prerequisites(entity.__class__)
        return self._add(entity)

    @abstractmethod
    def _add(self, entity: Union[ffd.Entity, List[ffd.Entity]]):
        pass

    def all(self, entity_type: Type[ffd.Entity], criteria: ffd.BinaryOp = None, limit: int = None, offset: int = None,
            sort: Tuple[Union[str, Tuple[str, bool]]] = None, raw: bool = False, count: bool = False):
        self._check_prerequisites(entity_type)
        return self._all(entity_type, criteria, limit, offset, raw=raw, count=count, sort=sort)

    @abstractmethod
    def _all(self, entity_type: Type[ffd.Entity], criteria: ffd.BinaryOp = None, limit: int = None, offset: int = None,
             sort: Tuple[Union[str, Tuple[str, bool]]] = None, raw: bool = False, count: bool = False):
        pass

    def find(self, uuid: str, entity_type: Type[ffd.Entity]):
        self._check_prerequisites(entity_type)
        return self._find(uuid, entity_type)

    @abstractmethod
    def _find(self, uuid: str, entity_type: Type[ffd.Entity]):
        pass

    def remove(self, entity: Union[ffd.Entity, List[ffd.Entity], Callable], force: bool = False):
        self._check_prerequisites(entity.__class__)
        deletions = []

        entities = entity
        if not isinstance(entity, list):
            entities = [entity]

        for entity in entities:
            if hasattr(entity, 'deleted_on') and not force:
                entity.deleted_on = datetime.now()
                self._update(entity)
            else:
                deletions.append(entity)

        if len(deletions) > 0:
            self._remove(deletions)

    @abstractmethod
    def _remove(self, entity: Union[ffd.Entity, List[ffd.Entity], Callable]):
        pass

    def update(self, entity: ffd.Entity):
        self._check_prerequisites(entity.__class__)
        if hasattr(entity, 'updated_on'):
            entity.updated_on = datetime.now()
        return self._update(entity)

    @abstractmethod
    def _update(self, entity: ffd.Entity):
        pass

    @abstractmethod
    def _ensure_connected(self):
        pass

    @abstractmethod
    def clear(self, entity: Type[ffd.Entity]):
        pass

    @abstractmethod
    def destroy(self, entity: Type[ffd.Entity]):
        pass

    def _check_prerequisites(self, entity: Type[ffd.Entity]):
        self._ensure_connected()

    @abstractmethod
    def _build_entity(self, entity: Type[ffd.Entity], data, raw: bool = False):
        pass

    def _get_relationships(self, entity: Type[ffd.Entity]):
        relationships = {}
        annotations_ = get_type_hints(entity)
        for k, v in annotations_.items():
            if k.startswith('_'):
                continue
            if isinstance(v, type) and issubclass(v, ffd.AggregateRoot):
                relationships[k] = {
                    'field_name': k,
                    'target': v,
                    'this_side': 'one',
                }
            elif ffd.is_type_hint(v):
                origin = ffd.get_origin(v)
                args = ffd.get_args(v)
                if origin is List and issubclass(args[0], ffd.AggregateRoot):
                    relationships[k] = {
                        'field_name': k,
                        'target': args[0],
                        'this_side': 'many',
                    }
        return relationships

    def _get_relationship(self, entity: Type[ffd.Entity], inverse_entity: Type[ffd.Entity]):
        relationships = self._get_relationships(entity)
        for k, v in relationships.items():
            if v['target'] == inverse_entity:
                return v

    def _load_relationships(self, entity: Type[ffd.Entity], data: dict):
        for k, v in self._get_relationships(entity).items():
            if v['this_side'] == 'one':
                data[k] = self._registry(v['target']).find(data[k])
            elif v['this_side'] == 'many':
                data[k] = list(self._registry(v['target']).filter(
                    lambda ee: getattr(ee, v['target'].id_name()).is_in(data[k])
                ))

        return data

    def _serialize_entity(self, entity: ffd.Entity):
        relationships = self._get_relationships(entity.__class__)
        if len(relationships.keys()) > 0:
            obj = entity.to_dict(force_all=True, skip=list(relationships.keys()))
            for k, v in relationships.items():
                if v['this_side'] == 'one':
                    obj[k] = getattr(entity, k).id_value()
                elif v['this_side'] == 'many':
                    obj[k] = list(map(lambda kk: kk.id_value(), getattr(entity, k)))
        else:
            obj = entity.to_dict(force_all=True)

        return self._serializer.serialize(obj)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Callable, Dict, Union

from firefly import domain as ffd
from firefly.domain.service.content_negotiation.content_converter import ContentConverter
from firefly.domain.service.logging import LoggerAware
from firefly.domain.service.messaging.middleware import Middleware


class ContentNegotiator(Middleware, LoggerAware):
    _converters: Dict[str, ContentConverter] = None
    _kernel: ffd.Kernel = None

    def __init__(self, converters: Dict[str, ContentConverter], logger: ffd.Logger):
        self._converters = converters
        self._logger = logger

    def __call__(self, message: Union[ffd.Envelope, ffd.Message], next_: Callable) -> ffd.Message:
        accept = None
        if isinstance(message, ffd.Envelope):
            accept = message.get_requested_content_type()

        response = next_(message)

        if accept is None:
            return response

        mimes = accept.split(',')
        ordered_mimes = []
        for mime_type in mimes:
            if ';' in mime_type:
                self.info(mime_type)
                parts = mime_type.split(';')
                mime_type = parts.pop(0)
                found = False
                for part in parts:
                    if part.startswith('q='):
                        ordered_mimes.append((mime_type, part.split('=')[1]))
                if not found:
                    ordered_mimes.append((mime_type, '1.0'))
            else:
                ordered_mimes.append((mime_type, '1.0'))
        ordered_mimes.sort(key=lambda i: i[1], reverse=True)

        for mime_type in ordered_mimes:
            mime_type = str(mime_type[0]).strip()
            if mime_type in self._converters and self._converters[mime_type].can_convert(message, response):
                return self._converters[mime_type].convert(message, response)

        return response



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import inspect
import keyword
import logging
import typing
from datetime import datetime, date

import firefly.domain as ffd

# __pragma__('skip')
from dataclasses import is_dataclass, fields
from abc import ABC
# __pragma__('noskip')
# __pragma__ ('ecom')
from dateparser import parse

"""?
from firefly.presentation.web.polyfills import is_dataclass, fields
?"""
# __pragma__ ('noecom')


keyword_list = keyword.kwlist
keyword_list.append('id')


def _fix_keywords(params: dict):
    if not isinstance(params, dict):
        return params

    for k, v in params.copy().items():
        if k in keyword_list:
            params[f'{k}_'] = v
    return params


class Void:
    pass


void = Void()


def _handle_type_hint(params: typing.Any, t: type, key: str = None, required: bool = False):
    # logging.debug('Processing type hint %s with params: %s, key: %s, required: %s', t, params, key, required)
    ret = {}

    origin = ffd.get_origin(t)
    args = ffd.get_args(t)

    if origin is typing.List:
        if key not in params:
            if required:
                raise ffd.MissingArgument(f'Missing argument {key} for type {t}')
            return []

        if ffd.is_type_hint(args[0]):
            if key is not None:
                ret[key] = list(map(lambda a: _handle_type_hint(a, args[0]), params[key]))
            else:
                ret = list(map(lambda a: _handle_type_hint(a, args[0]), params[key]))
        elif inspect.isclass(args[0]) and issubclass(args[0], ffd.ValueObject):
            try:
                if key is not None:
                    ret[key] = list(map(lambda a: _build_value_object(a, args[0], required), params[key]))
                else:
                    ret = list(map(lambda a: _build_value_object(a, args[0], required), params[key]))
            except TypeError:
                return
        else:
            ret[key] = params[key]

    elif origin is typing.Dict:
        if key not in params:
            if required:
                raise ffd.MissingArgument()
            return {}

        if ffd.is_type_hint(args[1]):
            ret[key] = {k: _handle_type_hint(v, args[1]) for k, v in params[key].items()}
        elif inspect.isclass(args[1]) and issubclass(args[1], ffd.ValueObject):
            ret[key] = {k: _build_value_object(v, args[1], required) for k, v in params[key].items()}
        else:
            ret[key] = params[key]

    elif origin is typing.Union:
        for arg in args:
            # logging.debug('Calling _handle_type_hint')
            r = _handle_type_hint(params, arg, key, required)
            # logging.debug('Response: %s', r)
            if r is not void:
                if key is not None:
                    ret[key] = r
                else:
                    ret = r
                break

    else:
        if inspect.isclass(t) and issubclass(t, ffd.ValueObject):
            if key in params:
                return _build_value_object(params[key], t, required)
            else:
                return _build_value_object(params, t, required)

        try:
            if key is not None:
                if key in params:
                    val = _check_special_types(params[key], t)
                    if val is not None:
                        return val

                    try:
                        if t(params[key]) == params[key]:
                            return params[key]
                    except ValueError:
                        pass
                    if params[key] is None:
                        return None
                elif str(t) == "<class 'NoneType'>":
                    return None

            elif key is None and t(params) == params:
                return params
        except (TypeError, KeyError):
            pass

    if ret == {}:
        return void

    return ret


def _build_value_object(obj, type_, required):
    try:
        if isinstance(obj, type_):
            return obj
    except TypeError:
        pass

    try:
        e = _generate_model(obj, type_)
        if e is False and required is True:
            raise ffd.MissingArgument()
        return e
    except ffd.MissingArgument:
        if required is False:
            return
        raise


def _generate_model(args: dict, model_type: type, strict: bool = False):
    subclasses = model_type.__subclasses__()
    if len(subclasses):
        for subclass in subclasses:
            try:
                return _generate_model(args, subclass, strict=True)
            except RuntimeError:
                continue

    entity_args = build_argument_list(args, model_type)
    if strict:
        for k in args.keys():
            if k not in entity_args:
                raise RuntimeError()

    return model_type(**entity_args)


def _check_special_types(value: typing.Any, type_: type):
    if type_ is datetime and isinstance(value, str):
        return parse(value).replace(tzinfo=None)
    elif type_ is date and isinstance(value, str):
        return parse(value).replace(tzinfo=None).date()


def build_argument_list(params: dict, obj: typing.Union[typing.Callable, type], strict: bool = True):
    # logging.debug('Building argument list for %s with params: %s, strict: %s', obj, params, strict)
    args = {}
    field_dict = {}
    is_dc = False
    params = _fix_keywords(params)

    if is_dataclass(obj):
        is_dc = True
        field_dict = {}
        # noinspection PyDataclass
        for field_ in fields(obj):
            field_dict[field_.name] = field_
        sig = inspect.signature(obj.__init__)
        types = typing.get_type_hints(obj)
    elif isinstance(obj, ffd.MetaAware):
        sig = inspect.signature(obj.__call__)
        types = typing.get_type_hints(obj.__call__)
    elif isinstance(obj, type):
        sig = inspect.signature(obj.__init__)
        types = typing.get_type_hints(obj.__init__)
    else:
        sig = inspect.signature(obj)
        try:
            types = typing.get_type_hints(obj)
        except NameError:
            types = obj.__annotations__

    has_kwargs = False
    for param in sig.parameters.values():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            has_kwargs = True
            # return params

    for name, param in sig.parameters.items():
        # logging.debug('Processing param %s', param)
        if name == 'self' or param.kind in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD):
            # logging.debug('Skipping...')
            continue

        required = False
        if is_dc:
            required = field_dict[name].metadata.get('required', False) is True
            try:
                d = field_dict[name].default_factory()
                if not isinstance(d, ffd.Empty):
                    required = False
            except (AttributeError, TypeError):
                pass
        elif param.default == inspect.Parameter.empty:
            required = True

        type_ = types[name] if name in types else None
        if name in params:
            val = _check_special_types(params[name], type_)
            if val is not None:
                params[name] = val

        if isinstance(type_, type) and issubclass(type_, ffd.ValueObject):
            if name in params and isinstance(params[name], type_):
                args[name] = params[name]
                continue

            try:
                nested = False
                e = None
                if name in params and isinstance(params[name], dict):
                    e = _generate_model(params[name], type_)
                    nested = True
                elif isinstance(params, dict):
                    e = _generate_model(copy_params(params, sig), type_)
            except ffd.MissingArgument:
                if required is False:
                    continue
                raise
            # TODO use factories where appropriate
            args[name] = e
            if nested:
                del params[name]

        elif ffd.is_type_hint(type_):
            if type_ is typing.Any:
                if name in params:
                    args[name] = params[name]
            else:
                if name in params:
                    parameter_args = _handle_type_hint({name: params[name]}, type_, key=name, required=required)
                elif isinstance(params, dict):
                    parameter_args = _handle_type_hint(copy_params(params, sig), type_, key=name, required=required)
                if parameter_args and not isinstance(parameter_args, Void):
                    args.update(parameter_args)

        elif isinstance(params, dict) and name in params:
            try:
                if params[name] is None:
                    args[name] = None
                elif isinstance(params[name], bytes):
                    args[name] = params[name]
                else:
                    args[name] = type_(params[name])
            except TypeError:
                args[name] = params[name]
        elif name.endswith('_') and name.rstrip('_') in params:
            # args[name] = params[name.rstrip('_')]
            try:
                sname = name.rstrip('_')
                if params[sname] is None:
                    args[name] = None
                elif isinstance(params[sname], bytes):
                    args[name] = params[sname]
                else:
                    args[name] = type_(params[sname])
            except TypeError:
                args[name] = params[name.rstrip('_')]
        elif required is True and strict:
            raise ffd.MissingArgument(f'Argument: {name} is required for object {obj}')

    if has_kwargs:
        for k, v in params.items():
            if k not in args:
                args[k] = v

    # logging.debug('Returning %s', args)
    return args


def copy_params(params: dict, sig: inspect.Signature):
    params_copy = params.copy()
    for n in sig.parameters.keys():
        if n in params_copy:
            del params_copy[n]
    return params_copy



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import inspect
from typing import Type

import firefly as ff
import firefly.infrastructure as ffi
import firefly_di as di
import inflection
from firefly.domain.entity.entity import Entity
from firefly.domain.service.core.application_service import ApplicationService
from firefly.domain.service.core.invoke_command import InvokeCommand
from firefly.domain.service.logging.logger import LoggerAware


class AutoGenerateAggregateApis(ApplicationService, LoggerAware):
    _context_map: ff.ContextMap = None
    _system_bus: ff.SystemBus = None
    _container: di.Container = None
    _command_resolving_middleware: ff.CommandResolvingMiddleware = None
    _event_resolving_middleware: ff.EventResolvingMiddleware = None
    _query_resolving_middleware: ff.QueryResolvingMiddleware = None

    def __call__(self, **kwargs):
        for context in self._context_map.contexts:
            for entity in context.entities:
                self._process_entity(context, entity)
            if 'extends' in context.config:
                base_context = self._context_map.get_context(context.config['extends'])
                for entity in base_context.entities:
                    entity._context = context.name
                    self._process_entity(context, entity)

    def _process_entity(self, context: ff.Context, entity: type):
        if not issubclass(entity, ff.AggregateRoot) or entity == ff.AggregateRoot:
            return

        self._create_crud_command_handlers(entity, context)
        self._create_query_handler(entity, context)
        self._register_entity_level_event_listeners(entity, context)
        for k in dir(entity):
            if k.startswith('_'):
                continue

            v = getattr(entity, k)
            if inspect.isfunction(v):
                fqn = self._create_invoke_command_handler(entity, context, k)
                if ff.has_meta(v) and ff.has_endpoints(v):
                    for endpoint in ff.get_endpoints(v):
                        endpoint.message = fqn
                        context.endpoints.append(endpoint)

    def _create_invoke_command_handler(self, entity: Type[Entity], context: ff.Context, method_name: str):
        command_name = f'{entity.__name__}::{inflection.camelize(method_name)}'

        class Invoke(InvokeCommand[entity]):
            pass

        Invoke.__name__ = f'Invoke{command_name}'
        fqn = f'{context.name}.{command_name}'

        self._command_resolving_middleware.add_command_handler(self._container.build(Invoke, method=method_name), fqn)
        context.command_handlers[Invoke] = fqn

        return fqn

    def _create_query_handler(self, entity: Type[Entity], context: ff.Context):
        query_name = inflection.pluralize(entity.__name__)

        class Query(ff.QueryService[entity]):
            pass

        Query.__name__ = query_name
        fqn = f'{context.name}.{query_name}'

        self._query_resolving_middleware.add_query_handler(self._container.build(Query), fqn)
        context.query_handlers[Query] = fqn

    def _create_crud_command_handlers(self, entity: Type[Entity], context: ff.Context):
        for action in ('Create', 'Delete', 'Update'):
            self._create_crud_command_handler(context, entity, action)

    def _create_crud_command_handler(self, context: ff.Context, entity, name_prefix):
        name = f'{name_prefix}Entity'
        base = getattr(ff, name)

        class Action(base[entity]):
            pass

        name = f'{name_prefix}{entity.__name__}'
        fqn = f'{context.name}.{name}'

        if not self._command_resolving_middleware.has_command_handler(fqn):
            Action.__name__ = name
            self._command_resolving_middleware.add_command_handler(self._container.build(Action), fqn)
            context.command_handlers[Action] = fqn

    def _register_entity_level_event_listeners(self, entity: Type[ff.AggregateRoot], context: ff.Context):
        if entity.get_create_on() is not None:
            self._register_crud_event_listener(
                entity.get_create_on(), f'{entity.get_class_context()}.Create{entity.__name__}', context
            )
        if entity.get_delete_on() is not None:
            self._register_crud_event_listener(
                entity.get_delete_on(), f'{entity.get_class_context()}.Delete{entity.__name__}', context
            )
        if entity.get_update_on() is not None:
            self._register_crud_event_listener(
                entity.get_update_on(), f'{entity.get_class_context()}.Update{entity.__name__}', context
            )

    def _register_crud_event_listener(self, event: ff.TypeOfEvent, command: str, context: ff.Context):
        class DoInvokeOn(ff.InvokeOn):
            pass

        self._event_resolving_middleware.add_event_listener(
            self._container.build(DoInvokeOn, command_name=command), event
        )
        if DoInvokeOn not in context.event_listeners:
            context.event_listeners[DoInvokeOn] = []

        context.event_listeners[DoInvokeOn].append(event)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import inspect
from pprint import pprint
from typing import Callable, Dict, Type, Union

import firefly.domain as ffd
from firefly.domain.entity.messaging.command import Command
from firefly.domain.service.logging.logger import LoggerAware
from firefly.domain.service.messaging.middleware import Middleware


class CommandResolvingMiddleware(Middleware, LoggerAware):
    _context_map: ffd.ContextMap = None
    _context: str = None
    _ff_environment: str = None

    def __init__(self, command_handlers: Dict[Type[Command], ffd.ApplicationService] = None):
        self._command_handlers = {}
        self._initialized = False
        if command_handlers is not None:
            for command, handler in command_handlers.items():
                self._command_handlers[command.get_fqn() if not isinstance(command, str) else command] = handler

    def _initialize(self):
        for command, handler in self._command_handlers.items():
            if inspect.isclass(handler):
                self._command_handlers[command] = \
                    self._context_map.get_context(handler.get_class_context()).container.build(handler)
        self._initialized = True

    def __call__(self, message: ffd.Message, next_: Callable) -> ffd.Message:
        if not self._initialized:
            self._initialize()

        not_for_this_context = message.get_context() != 'firefly' and \
                               self._context_map.get_context(message.get_context()) is None
        is_async = hasattr(message, '_async') and getattr(message, '_async') is True

        if not_for_this_context or is_async:
            return self._transfer_message(message)

        args = message.to_dict()
        args['_message'] = message

        if str(message) not in self._command_handlers:
            raise ffd.ConfigurationError(f'No command handler registered for {message}')

        service = self._command_handlers[str(message)]

        parsed_args = ffd.build_argument_list(args, service)
        self.debug('Calling service %s with arguments: %s', service, parsed_args)
        return service(**parsed_args)

    def _transfer_message(self, message: ffd.Message):
        return self._context_map.get_context(self._context).container.message_transport.invoke(message)

    def has_command_handler(self, handler: str):
        return handler in self._command_handlers

    def add_command_handler(self, handler: Union[ffd.ApplicationService, Type[ffd.ApplicationService]],
                            command: Union[Type[Command], str]):
        if inspect.isclass(handler):
            handler = self._context_map.get_context(handler.get_class_context()).container.build(handler)
        if inspect.isclass(command):
            command = command.get_fqn()
        self._command_handlers[command.get_fqn() if not isinstance(command, str) else command] = handler



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

# __pragma__('skip')
from abc import ABC
# __pragma__('noskip')
# __pragma__ ('ecom')
"""?
from firefly.presentation.web.polyfills import ABC
?"""
# __pragma__ ('noecom')


class ContextAware(ABC):
    @classmethod
    def get_class_context(cls):
        if hasattr(cls, '_context') and cls._context is not None:
            return cls._context

        parts = cls.__module__.split('.')
        # KLUDGE For integration / acceptance testing
        return parts[0] if parts[0] != 'firefly_test' else parts[1]

    @classmethod
    def set_class_context(cls, context_name: str):
        cls._context = context_name

    @classmethod
    def get_fqn(cls):
        if hasattr(cls, '_context') and cls._context is not None:
            context = cls._context
        else:
            context = cls.get_class_context()
        return f'{context}.{cls.__name__}'



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import inspect
from typing import Callable, Dict, Type, Union

import firefly.domain as ffd
from firefly.domain.entity.messaging.query import Query
from firefly.domain.service.logging.logger import LoggerAware
from firefly.domain.service.messaging.middleware import Middleware


class QueryResolvingMiddleware(Middleware, LoggerAware):
    _context_map: ffd.ContextMap = None
    _context: str = None
    _ff_environment: str = None

    def __init__(self, query_handlers: Dict[ffd.ApplicationService, Type[Query]] = None):
        self._initialized = False
        self._query_handlers = query_handlers or {}

    def _initialize(self):
        for query, handler in self._query_handlers.items():
            if inspect.isclass(handler):
                self._query_handlers[query] = \
                    self._context_map.get_context(handler.get_class_context()).container.build(handler)
        self._initialized = True

    def __call__(self, message: ffd.Message, next_: Callable) -> ffd.Message:
        if not self._initialized:
            self._initialize()

        if message.get_context() != 'firefly' and self._context_map.get_context(message.get_context()) is None:
            return self._transfer_message(message)

        args = message.to_dict()
        args['_message'] = message
        for service, query_type in self._query_handlers.items():
            if message.is_this(query_type):
                parsed_args = ffd.build_argument_list(args, service)
                self.debug('Calling service with arguments: %s', parsed_args)
                return service(**parsed_args)
        raise ffd.ConfigurationError(f'No query handler registered for {message}')

    def _transfer_message(self, message: ffd.Message):
        return self._context_map.get_context(self._context).container.message_transport.request(message)

    def add_query_handler(self, handler: Union[ffd.ApplicationService, Type[ffd.ApplicationService]],
                          query: Union[Type[Query], str]):
        if inspect.isclass(handler):
            handler = self._context_map.get_context(handler.get_class_context()).container.build(handler)
        if inspect.isclass(query):
            query = query.get_fqn()

        self._query_handlers[handler] = query



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

# __pragma__('skip')
from abc import ABCMeta
from dataclasses import dataclass
# __pragma__('noskip')
# __pragma__ ('ecom')
"""?
from firefly.presentation.web.polyfills import ABCMeta, ABC, dataclass
?"""
# __pragma__ ('noecom')

import firefly.domain as ffd


class ValueObjectMeta(ABCMeta):
    def __new__(mcs, name, bases, dct, **kwargs):
        # __pragma__('skip')
        ret = super().__new__(mcs, name, bases, dct)
        # __pragma__('noskip')
        # __pragma__ ('ecom')
        """?
        ret = type.__new__(type, name, bases, dct)
        ?"""
        # __pragma__ ('noecom')
        ret = dataclass(ret, eq=False)

        return ret



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Callable

import firefly.domain as ffd

from .registry import Registry
from ..service.logging.logger import LoggerAware
from ..service.messaging.middleware import Middleware
from ..service.messaging.system_bus import SystemBusAware


class TransactionHandlingMiddleware(Middleware, LoggerAware, SystemBusAware):
    _registry: Registry = None

    def __init__(self):
        self._level = 0
        self._event_buffer = []

    def reset_level(self):
        self._level = 0

    def __call__(self, message: ffd.Message, next_: Callable) -> ffd.Message:
        while True:
            try:
                if self._level == 0:
                    self.debug('Level 0 - Resetting repositories')
                    self.debug(message)
                    self._reset()
                elif self._level > 0:
                    if isinstance(message, ffd.Event):
                        self.debug('Buffering message')
                        self._event_buffer.append(message)
                        return message

                self._level += 1
                self.debug('Level incremented: %d', self._level)
                ret = next_(message)
                self._level -= 1
                if self._level < 0:
                    self._level = 0
                self.debug('Level decremented: %d', self._level)
                if self._level == 0:
                    self.debug('Level 0 - Committing changes')
                    self._commit()
                return ret
            except ffd.ConcurrentUpdateDetected:
                self.info('Concurrent update detected. Retrying the operation.')
                self.reset_level()
                self._reset()
            except Exception as e:
                self.exception(str(e))
                self._level -= 1
                if self._level < 0:
                    self._level = 0
                self.debug('Level decremented: %d', self._level)
                if self._level == 0:
                    self.debug('Level 0 - Resetting repositories')
                    self._reset()
                raise

    def _reset(self):
        for repository in self._registry.get_repositories():
            repository.reset()
        self._event_buffer = []

    def _commit(self):
        for repository in self._registry.get_repositories():
            self.debug('Committing repository %s', repository)
            repository.commit()
        self.debug('Dispatching events %s', [{e: e.to_dict() for e in self._event_buffer}])
        list(map(lambda e: self.dispatch(e), self._event_buffer))



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

import setuptools


setuptools.setup(
    name='firefly-framework',
    version='1.1.59',
    author="JD Williams",
    author_email="me@jdwilliams.xyz",
    description="A SOA framework for Python",
    # long_description_content_type="text/markdown",
    url="https://github.com/firefly-framework/firefly-framework",
    data_files=[('firefly_config', ['firefly.yml'])],
    entry_points={
        'console_scripts': ['firefly=firefly.presentation.cli:main'],
        'pytest11': ['firefly=firefly.plugins.pytest']
    },
    install_requires=[
        'aiohttp>=3.5.4',
        'aiohttp-cors>=0.7.0',
        'dateparser>=0.7.4',
        'dirsync>=2.2.3',
        'inflection>=0.3.1',
        'Jinja2==2.11.3',
        'jinjasql>=0.1.8',
        'python-dateutil>=2.8.1',
        'python-dotenv>=0.10.3',
        'pyyaml>=5.1.1',
        'routes>=2.4.1',
        'terminaltables>=3.1.0',
        'websockets>=8.0.2',
        'firefly-dependency-injection>=0.1',
    ],
    extras_require={
        'OpenApi Generation': ['apispec>=3.3.0', 'docstring_parser>=0.7.1'],
        'Frontend Support': ['transcrypt>=3.7.16']
    },
    packages=setuptools.PEP420PackageFinder.find('src'),
    package_dir={'': 'src'},
    package_data={'firefly': [
        'infrastructure/repository/rdb_storage_interfaces/templates/**/*.sql',
    ]},
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ]
)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

import uuid
from datetime import datetime
from multiprocessing import Pool
from pprint import pprint

from firefly_test.iam import Scope, Role, User
from firefly_test.todo import TodoList, User as TodoUser, Task

id_ = str(uuid.uuid4())


def test_mutability(users):
    subset = users.filter(lambda u: u.name == 'Bob Loblaw')

    assert len(users) == 4
    assert len(subset) == 1
    assert subset[0].name == 'Bob Loblaw'
    assert len(users.filter(lambda u: u.email.is_in(('davante@adams.com', 'bob@loblaw.com')))) == 2
    assert len(users) == 4

    half = users.sort(lambda u: u.name)[2:]
    assert len(half) == 2
    assert half[0].name == 'David Johnson'
    assert half[1].name == 'John Doe'

    bob = users.filter(lambda u: u.name == 'Bob Loblaw')[0]
    bob.email = 'foo@bar.com'
    users.commit()
    assert users.find(lambda u: u.name == 'Bob Loblaw').email == 'foo@bar.com'

    subset = users.filter(lambda u: u.email.is_in(('davante@adams.com', 'bob@loblaw.com'))).sort(lambda u: u.email)
    subset.remove(subset[0])
    users.commit()
    assert len(users) == 3


def test_basic_crud_operations(todos):
    todos.migrate_schema()

    todos.append(TodoList(id=id_, user=TodoUser(name='Bob')))
    todos.commit()
    todos.reset()

    assert len(todos) == 1
    todo: TodoList = todos.find(id_)
    assert todo is not None
    assert todo.user.name == 'Bob'

    todo.tasks.append(Task(name='Task 1', due_date=datetime.now()))
    todos.commit()
    todos.reset()

    assert len(todos) == 1
    todo: TodoList = todos.find(id_)
    assert len(todo.tasks) == 1

    todo.user.name = 'Phillip'
    todos.commit()
    todos.reset()

    assert len(todos) == 1
    todo: TodoList = todos.find(id_)
    assert todo.user.name == 'Phillip'

    todos.remove(todo)
    todos.commit()
    todos.reset()

    assert len(todos) == 0
    assert todos.find(id_) is None


def test_aggregate_associations(users):
    bob = users.find(lambda u: u.name == 'Bob Loblaw')

    assert bob.roles[0].name == 'Admin User'
    assert bob.roles[0].scopes[0].id == 'firefly.admin'


def test_pagination(users):
    test = users.sort(lambda u: u.name)[0:1]

    assert len(test) == 2
    assert test[0].name == 'Bob Loblaw'
    assert test[1].name == 'Davante Adams'

    test = users.sort(lambda u: u.name)[1:2]

    assert len(test) == 2
    assert test[0].name == 'Davante Adams'
    assert test[1].name == 'David Johnson'

    test = users.sort(lambda u: u.name).filter(lambda u: u.name.is_in(('Davante Adams', 'David Johnson')))
    assert test[0].name == 'Davante Adams'
    assert test[1].name == 'David Johnson'


def test_list_expansion(users):
    my_users = users.filter(lambda u: u.name.is_in(['Bob Loblaw', 'Davante Adams']))

    emails = list(map(lambda u: u.email, my_users))
    emails.sort()
    assert emails == ['bob@loblaw.com', 'davante@adams.com']

    my_users = list(filter(lambda u: u.name == 'Bob Loblaw', users))
    assert len(my_users) == 1
    assert my_users[0].name == 'Bob Loblaw'


def iam_fixtures(users, roles, scopes):
    my_scopes = [
        Scope(id='firefly.admin'),
        Scope(id='firefly.read'),
        Scope(id='firefly.write'),
    ]
    list(map(lambda s: scopes.append(s), my_scopes))
    scopes.commit()

    my_roles = [
        Role(name='Anonymous User', scopes=[my_scopes[1]]),
        Role(name='Admin User', scopes=[my_scopes[0]]),
        Role(name='Regular User', scopes=[my_scopes[1], my_scopes[2]]),
    ]
    list(map(lambda r: roles.append(r), my_roles))
    roles.commit()

    users.append(User(name='John Doe', email='john@doe.com', roles=[my_roles[2]]))
    users.append(User(name='Bob Loblaw', email='bob@loblaw.com', roles=[my_roles[1]]))
    users.append(User(name='David Johnson', email='david@johnson.com', roles=[my_roles[1], my_roles[2]]))
    users.append(User(name='Davante Adams', email='davante@adams.com', roles=[my_roles[0]]))
    users.commit()

    users.reset()
    roles.reset()
    scopes.reset()



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Union, Any

from ..entity import Entity, id_, dict_, required


class Envelope(Entity):
    id: str = id_()
    headers: dict = dict_()
    raw_request: dict = dict_()
    body: Any = required()

    @classmethod
    def wrap(cls, body: Any) -> Envelope:
        return cls(body=body)

    def unwrap(self):
        return self.body

    def set_requested_content_type(self, type_: str):
        self.headers['accept'] = type_
        return self

    def get_requested_content_type(self):
        return self.headers.get('accept')

    def set_raw_request(self, source_headers: Any):
        self.raw_request = source_headers
        return self

    def get_raw_request(self):
        return self.raw_request

    def set_range(self, lower: int, upper: int, total: int, unit: str = None):
        self.headers['range'] = {
            'lower': lower,
            'upper': upper,
            'total': total,
        }
        if unit is not None:
            self.headers['range']['unit'] = unit
        return self

    def get_range(self):
        return self.headers.get('range')

    def add_forwarding_address(self, location: str):
        self.headers['location'] = location
        return self



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Union, Type

import firefly.domain as ffd
from firefly.domain.meta.firefly_type import FireflyType
from firefly.domain.meta.meta_aware import MetaAware

from ..logging.logger import LoggerAware
from ..messaging.system_bus import SystemBusAware


class ApplicationService(FireflyType, MetaAware, ABC, SystemBusAware, LoggerAware):
    _event_buffer: ffd.EventBuffer = None
    _kernel: ffd.Kernel = None

    @abstractmethod
    def __call__(self, **kwargs):
        pass

    def _buffer_events(self, events):
        if isinstance(events, list):
            for event in events:
                if isinstance(event, (ffd.Event, tuple)):
                    self._event_buffer.append(event)
        elif isinstance(events, ffd.Event) or (isinstance(events, tuple) and len(events) == 2):
            self._event_buffer.append(events)

        return events

    @classmethod
    def get_arguments(cls) -> dict:
        return ffd.get_arguments(cls.__call__)

    @classmethod
    def locate_message(cls, message: Union[str, Type[ffd.Message]]):
        if cls.is_event_listener():
            for event in cls.get_events():
                if message == event:
                    return event, 'event'
        if cls.is_command_handler():
            if message == cls.get_command():
                return cls.get_command(), 'command'
        if cls.is_query_handler():
            if message == cls.get_query():
                return cls.get_query(), 'query'



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

import importlib
import inspect

import firefly.domain as ffd
import firefly.infrastructure as ffi
import inflection


@ffd.on(ffd.DomainEntitiesLoaded)
class ConfigureStorage(ffd.ApplicationService):
    _config: ffd.Configuration = None
    _context_map: ffd.ContextMap = None
    _registry: ffd.Registry = None
    _serializer: ffd.Serializer = None
    _rdb_storage_interface_registry: ffi.RdbStorageInterfaceRegistry = None

    def __init__(self):
        self._connection_factories = {}
        self._repository_factories = {}

    def __call__(self, **kwargs):
        self._load_factories()
        connections = {}
        factories = {}

        for context in self._context_map.contexts:
            storage = context.config.get('storage', {})
            if 'services' in storage:
                for name, config in storage.get('services').items():
                    config = config or {}
                    if name not in self._connection_factories and config.get('type') not in self._connection_factories:
                        raise ffd.ConfigurationError(f"No ConnectionFactory configured for '{name}'")
                    key = name if name in self._connection_factories else config.get('type')
                    connections[name] = context.container.build(self._connection_factories[key])(
                        **(config.get('connection') or {})
                    )

        for context in self._context_map.contexts:
            storage = context.config.get('storage', {})
            if 'services' in storage:
                for name, config in storage.get('services').items():
                    config = config or {}
                    if name not in self._repository_factories and config.get('type') not in self._repository_factories:
                        raise ffd.ConfigurationError(f"No RepositoryFactory configured for '{name}'")
                    key = name if name in self._repository_factories else config.get('type')
                    factory = context.container.autowire(self._repository_factories[key])
                    try:
                        factories[name] = factory(connections[name], **(config.get('repository') or {}))
                    except TypeError as e:
                        if '__init__() takes exactly one argument' in str(e):
                            raise ffd.FrameworkError(f"{factory.__name__}.__init__() must take a connection as "
                                                     f"the first argument")
                        raise e

        for context in self._context_map.contexts:
            if context.name == 'firefly':
                continue
            storage = context.config.get('storage', {})
            registered_aggregates = []
            if 'aggregates' in storage:
                for entity, service in storage.get('aggregates').items():
                    if not entity.startswith(context.name):
                        entity = f'{context.name}.{entity}'
                    entity = ffd.load_class(entity)
                    registered_aggregates.append(entity)
                    self._registry.register_factory(entity, factories[service])
            if 'default' in storage:
                for entity in context.entities:
                    if issubclass(entity, ffd.AggregateRoot) and entity is not ffd.AggregateRoot \
                            and entity not in registered_aggregates:
                        self._registry.register_factory(entity, factories[storage.get('default')])

        # TODO Get persistence working in these core services.
        # self._registry(ffd.ContextMap).add(self._context_map)

        self.dispatch(ffd.StorageConfigured())

    def _load_factories(self):
        for context in self._context_map.contexts:
            module_name = context.config.get('infrastructure_module', '{}.infrastructure')
            try:
                module = importlib.import_module(module_name.format(context.name))
            except ModuleNotFoundError:
                continue

            for k, v in module.__dict__.items():
                if not inspect.isclass(v):
                    continue
                if issubclass(v, ffd.ConnectionFactory):
                    self._connection_factories[inflection.underscore(v.__name__.replace('ConnectionFactory', ''))] = v
                elif issubclass(v, ffd.RepositoryFactory):
                    self._repository_factories[inflection.underscore(v.__name__.replace('RepositoryFactory', ''))] = v
                elif issubclass(v, ffi.RdbStorageInterface) and v is not ffi.RdbStorageInterface:
                    name = inflection.underscore(k.replace('StorageInterface', ''))
                    self._rdb_storage_interface_registry.add(name, v)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

# __pragma__('skip')
from dataclasses import make_dataclass, fields, is_dataclass, asdict
from typing import Union, TypeVar, Type, Tuple, get_type_hints
# __pragma__('noskip')
# __pragma__ ('ecom')
"""?
from firefly.presentation.web.polyfills import make_dataclass, fields, is_dataclass, asdict, Union, TypeVar, Type, Tuple, get_type_hints
?"""
# __pragma__ ('noecom')

import firefly.domain as ffd

from firefly.domain.entity.messaging.message import Message

# __pragma__('kwargs')


class MessageFactory:
    @staticmethod
    def convert_type(message: Message, new_name: str, new_base: Union[Message, Tuple[Message]]):
        if not is_dataclass(message):
            raise ffd.FrameworkError('message must be a dataclass')

        types = get_type_hints(message.__class__)
        message_fields = []
        for field_ in fields(message):
            message_fields.append((field_.name, types[field_.name], field_))

        if not isinstance(new_base, tuple):
            new_base = (new_base,)

        cls = make_dataclass(new_name, fields=message_fields, bases=new_base, eq=False, repr=False)

        return cls(**asdict(message))

    def event(self, name: str, data: dict = None):
        return self._build(name, data or {}, (ffd.Event,))

    def event_class(self, name: str, fields_: dict = None):
        return self._build_message_class(name, fields_, (ffd.Event,))

    def command(self, name: str, data: dict = None):
        return self._build(name, data or {}, (ffd.Command,))

    def command_class(self, name: str, fields_: dict = None):
        return self._build_message_class(name, fields_, (ffd.Command,))

    def query(self, name: str, criteria: ffd.BinaryOp = None, data: dict = None):
        data = data or {}
        if criteria is not None:
            data['criteria'] = criteria.to_dict()
        return self._build(name, data, (ffd.Query,))

    def query_class(self, name: str, fields_: dict = None):
        return self._build_message_class(name, fields_, (ffd.Command,))

    def _build(self, name: str, data: dict, bases: tuple):
        # __pragma__ ('ecom')
        """?
        return make_dataclass(name, data, bases=bases)(**data)
        ?"""
        # __pragma__ ('noecom')

        return self._build_message_class(name, {k: type(v) for k, v in data.items()}, bases)(**data)

    @staticmethod
    def _build_message_class(name: str, fields_: dict, bases: tuple):
        # __pragma__('skip')
        context = None
        if '.' in name:
            context, name = name.split('.')
        ret = ffd.MessageMeta.__new__(ffd.MessageMeta, name, bases, fields_, fields_=fields_, annotations_=fields_)
        if context is not None:
            ret._context = context
        return ret
        # __pragma__('noskip')



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

"""A clone of the default copy.deepcopy that doesn't handle cyclic
structures or complex types except for dicts and lists. This is
because gyp copies so large structure that small copy overhead ends up
taking seconds in a project the size of Chromium."""

class Error(Exception):
  pass

__all__ = ["Error", "deepcopy"]

def deepcopy(x):
  """Deep copy operation on gyp objects such as strings, ints, dicts
  and lists. More than twice as fast as copy.deepcopy but much less
  generic."""

  try:
    return _deepcopy_dispatch[type(x)](x)
  except KeyError:
    raise Error('Unsupported type %s for deepcopy. Use copy.deepcopy ' +
                'or expand simple_copy support.' % type(x))

_deepcopy_dispatch = d = {}

def _deepcopy_atomic(x):
  return x

for x in (type(None), int, long, float,
          bool, str, unicode, type):
  d[x] = _deepcopy_atomic

def _deepcopy_list(x):
  return [deepcopy(a) for a in x]
d[list] = _deepcopy_list

def _deepcopy_dict(x):
  y = {}
  for key, value in x.iteritems():
    y[deepcopy(key)] = deepcopy(value)
  return y
d[dict] = _deepcopy_dict

del d



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

# __pragma__('js', '{}', "export const m = require('mithril');")
# __pragma__('js', '{}', "export const Stream = require('mithril/stream');")
# __pragma__('skip')


class Route:
    prefix = None

    def __call__(self, root, default_route, routes):
        pass

    def set(self, path):
        pass

    def get(self):
        pass


class M:
    def __init__(self):
        self.route = Route()

    def __call__(self, selector, attrs, children=None):
        pass

    def render(self):
        pass

    def mount(self, element, component):
        pass

    def request(self, options):
        pass

    def jsonp(self, options):
        pass

    def parseQueryString(self, querystring):
        pass

    def buildQueryString(self, obj):
        pass

    def trust(self, html):
        pass

    def redraw(self):
        pass


m = M()


class Stream:
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        pass

    def __getattr__(self, item):
        pass



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Type

import firefly.domain as ffd
import firefly.infrastructure as ffi
import firefly_di as di
from firefly import Repository
from firefly.domain.repository.repository_factory import E


class DocumentRepositoryFactory(ffd.RepositoryFactory):
    _context_map: ffd.ContextMap = None
    _container: di.Container = None

    def __init__(self, interface: ffi.DocumentStorageInterface):
        self._interface = interface
        self._cache = {}
        self._default_interface = None

    def __call__(self, entity: Type[E]) -> Repository:
        if entity not in self._cache:
            class LocalRepository(ffi.DocumentRepository[entity]):
                pass
            LocalRepository.__name__ = f'{entity.__name__}Repository'
            params = self._get_repository_arguments(entity)
            params['interface'] = self._interface
            self._cache[entity] = self._container.build(LocalRepository, **params)

        return self._cache[entity]

    def _get_repository_arguments(self, entity: Type[E]):
        context = self._context_map.get_context(entity.get_class_context())
        if 'aggregates' not in context.config.get('storage', {}):
            return {}
        aggregates = context.config.get('storage').get('aggregates')
        if entity.__name__ in aggregates and isinstance(aggregates.get(entity.__name__), dict):
            return aggregates.get(entity.__name__)
        if entity.get_fqn() in aggregates and isinstance(aggregates.get(entity.get_fqn()), dict):
            return aggregates.get(entity.get_fqn())

        return {}



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import TypeVar, Generic

import firefly.domain as ffd
import inflection

from ..core.application_service import ApplicationService
from ...entity.aggregate_root import AggregateRoot
from ...value_object.generic_base import GenericBase

T = TypeVar('T', bound=AggregateRoot)


class InvokeCommand(Generic[T], GenericBase, ApplicationService):
    _registry: ffd.Registry = None

    def __init__(self, method: str):
        self._method = method

    def __call__(self, **kwargs):
        try:
            aggregate = self._registry(self._type()).find(kwargs[self._aggregate_name_snake()])
        except KeyError:
            raise ffd.MissingArgument(self._aggregate_name_snake())

        method = getattr(aggregate, self._method)
        return self._buffer_events(
            method(**ffd.build_argument_list(kwargs, method))
        )

    def _aggregate_name_snake(self):
        return inflection.underscore(self._type().__name__)



#  Copyright (c) 2019 JD Williams
#
#  This file is part of Firefly, a Python SOA framework built by JD Williams. Firefly is free software; you can
#  redistribute it and/or modify it under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 3 of the License, or (at your option) any later version.
#
#  Firefly is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#  Public License for more details. You should have received a copy of the GNU Lesser General Public
#  License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  You should have received a copy of the GNU General Public License along with Firefly. If not, see
#  <http://www.gnu.org/licenses/>.
import os

import firefly.domain as ffd


@ffd.cli('firefly generate project', alias={'path': ['-p'], 'name': ['-n']})
class InitializeProject(ffd.ApplicationService):
    def __call__(self, name: str = None, path: str = None):
        if name is None:
            name = input('Project name: ')

        if path is None:
            path = input('Path (.): ')
            if path is None or str(path).strip() == '':
                path = '.'

        os.chdir(path)
        self._generate_files(name)

    @staticmethod
    def _generate_files(name: str):
        with open('./README.md', 'w') as fp:
            fp.write(readme(name))
        with open('./setup.py', 'w') as fp:
            fp.write(setup(name))
        with open('./firefly.yml', 'w') as fp:
            fp.write(firefly_config(name))
        with open('./pytest.ini', 'w') as fp:
            fp.write(pytest_ini())

        try:
            os.mkdir('src')
            os.mkdir(f'src/{name}')
            os.mkdir(f'src/{name}/domain')
            os.mkdir(f'src/{name}/infrastructure')
            os.mkdir(f'src/{name}/application')
            os.mkdir(f'src/{name}/presentation')
        except FileExistsError:
            pass


def setup(name: str):
    return f"""
import setuptools
from setuptools.command.develop import develop
from setuptools.command.install import install

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name='{name}',
    version='0.1',
    author="",
    author_email="",
    description="Put project description here.",
    long_description=long_description,
    url="",
    entry_points={{
        'console_scripts': ['firefly=firefly.presentation.cli:main']
    }},
    install_requires=[
        'firefly-dependency-injection>=0.1',
    ],
    packages=setuptools.PEP420PackageFinder.find('src'),
    package_dir={{'': 'src'}},
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ]
)
""".lstrip()


def readme(name: str):
    return f"""
{name} README
""".lstrip()


def firefly_config(name):
    return f"""
contexts:
  firefly: ~
  {name}: ~
""".lstrip()


def pytest_ini():
    return """
[pytest]
log_cli = True
log_cli_level = DEBUG
""".lstrip()
